﻿''' <summary>
''' Car
''' </summary>
''' <remarks>
''' Eine Klasse ist ein Software-Baustein, der durch seinen Zustand (state) und sein Verhalten (behavior) beschrieben wird. 
''' Der Zustand ist in Daten gespeichert, das Verhalten wird in Methoden, Events und Exceptions abgebildet. 
''' Eine Klasse ist wie ein Bauplan, der mehrfach verwendet werden kann, um reale Dinge zu bauen. 
''' </remarks>
Public Class Car
    Implements IDisposable

#Region "member variables"
    Private ReadOnly mManufacturer As CarManufacturer
    Private ReadOnly mModel As String
    Private ReadOnly mLicensePlate As String
    Private mFuel As Double                     ' in l
    Private ReadOnly mConsumption As Double     ' in l/100 km
    Private ReadOnly mTopSpeed As Integer       ' in km/h
    Private mCurrentSpeed As Integer            ' in km/h
    Private mCurrentPosition As GPSPosition     ' GPS position
    Private mTotalDistanceDriven As Integer     ' in km
#End Region

#Region "events"
    Public Event NoFuel()   'car needs more gas
    Public Event Speed(currentSpeed As Integer)    'car reached this speed
#End Region

#Region "constructors"
    Public Sub New(manufacturer As CarManufacturer, model As String, licensePlate As String, topSpeed As Integer, consumption As Double)
        'check params:
        If String.IsNullOrWhiteSpace(model) Then Throw New Exception("Model must be defined.")
        If String.IsNullOrWhiteSpace(licensePlate) Then Throw New Exception("LicensePlate must be defined.")
        If (topSpeed <= 0) Then Throw New Exception("TopSpeed must be > 0")
        If (topSpeed > 400) Then Throw New Exception("TopSpeed is unreasonably high.")
        If (consumption <= 0) Then Throw New Exception("Consumption must be > 0")
        If (consumption > 40.0) Then Throw New Exception("Consumption is unreasonably high.")

        'store params:
        mManufacturer = manufacturer
        mModel = model
        mLicensePlate = licensePlate
        mTopSpeed = topSpeed
        mConsumption = consumption

        'init remaining state:
        mFuel = 30.0            'a new car comes with a few litres of gas
    End Sub
#End Region

#Region "properties"
    Public ReadOnly Property Manufacturer() As CarManufacturer
        Get
            Return mManufacturer
        End Get
    End Property

    Public ReadOnly Property Model() As String
        Get
            Return mModel
        End Get
    End Property

    Public ReadOnly Property LicensePlate() As String
        Get
            Return mLicensePlate
        End Get
    End Property

    Public ReadOnly Property TopSpeed() As Double
        Get
            Return mTopSpeed
        End Get
    End Property

    Public ReadOnly Property Consumption() As Double
        Get
            Return mConsumption
        End Get
    End Property

    Public ReadOnly Property CurrentSpeed() As Integer
        Get
            Return mCurrentSpeed
        End Get
    End Property

    Public Property Fuel() As Double
        Get
            Return mFuel
        End Get
        Set(value As Double)        ' fill up the tank
            If (value <= 0) Then Throw New Exception("You can only add a positive amount of fuel!")
            mFuel = value
        End Set
    End Property
#End Region

#Region "methods"
    Public Sub Accelerate(targetSpeed As Integer)
        If (targetSpeed < 0) Then Throw New Exception("Target speed must be >= 0!")
        mCurrentSpeed = targetSpeed     'car reaches new speed immediately
        mFuel = mFuel - 10              'burning some fuel
        If (mFuel <= 0) Then            'any fuel left?
            mFuel = 0                   'bad luck - we have no more fuel left
            mCurrentSpeed = 0           'so the car stops
            RaiseEvent NoFuel()         'inform others we need gas
        Else
            mTotalDistanceDriven += 30  'this drives us another 30 km
        End If
        RaiseEvent Speed(mCurrentSpeed)         'inform others about new speed
    End Sub

    Public Overrides Function ToString() As String
        Dim s As String
        's = mManufacturer.tostring & " " & mModel & ", " & mLicensePlate & ", Top speed: " & mTopSpeed & " km/h, Consumption: " & mConsumption & " l/100 km."

        'this can also be written as (using Me and line continuation):
        s = Me.Manufacturer.ToString & " " & Me.Model & ", " & Me.LicensePlate & _
            ", Top speed: " & Me.TopSpeed & " km/h, Consumption: " & Me.Consumption & " l/100 km."

        s = s & " Total distance driven: " & Me.mTotalDistanceDriven & " km."
        Return s
    End Function
#End Region

#Region "destructor"

#Region "IDisposable Support"
    Private disposedValue As Boolean ' To detect redundant calls

    ' IDisposable
    Protected Overridable Sub Dispose(disposing As Boolean)
        If Not Me.disposedValue Then
            If disposing Then
                ' TODO: dispose managed state (managed objects).
                mCurrentSpeed = 0   'halt the car
            End If

            ' TODO: free unmanaged resources (unmanaged objects) and override Finalize() below.
            ' TODO: set large fields to null.
        End If
        Me.disposedValue = True
    End Sub

    ' TODO: override Finalize() only if Dispose(ByVal disposing As Boolean) above has code to free unmanaged resources.
    'Protected Overrides Sub Finalize()
    '    ' Do not change this code.  Put cleanup code in Dispose(ByVal disposing As Boolean) above.
    '    Dispose(False)
    '    MyBase.Finalize()
    'End Sub

    ' This code added by Visual Basic to correctly implement the disposable pattern.
    Public Sub Dispose() Implements IDisposable.Dispose
        ' Do not change this code.  Put cleanup code in Dispose(disposing As Boolean) above.
        Dispose(True)
        GC.SuppressFinalize(Me)
    End Sub
#End Region
#End Region

    ''' <summary>
    ''' List of all car manufacturers
    ''' </summary>
    ''' <remarks></remarks>
    Public Enum CarManufacturer
        Ferrari
        Lamborghini
        Maserati
    End Enum
End Class


